"""
This file contains the File class, which is used to represent a file in the file system.
"""

from dataclasses import dataclass, field
import datetime
import logging
from pathlib import Path
import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .directory import Directory

logger = logging.getLogger(__name__)


@dataclass
class File:
    """
    A File object represents a file in the file system.

    Attributes:
        path (Path): The path to the file.
        parent (Directory): The parent directory of the file.

    Properties:
        file_metadata (dict): The metadata of the file.
        name (str): The name of the file.
        extension (str): The extension of the file.
        size (int): The size of the file in bytes.
        created (datetime): The date and time the file was created.
        modified (datetime): The date and time the file was last modified.

    Methods:
        get_file_metadata: Returns the metadata of the file.
        as_record: Returns the file metadata as a record.
    """

    DATE_FORMAT = "%d %B, %Y %H:%M"

    path: Path
    parent: "Directory" = field(default=None, repr=False)

    def __post_init__(self):
        self._metadata = None
        logger.debug("Created file object '%s'", self.path)

    @property
    def file_metadata(self) -> dict:
        """
        The metadata of the file as a dictionary.

        Returns:
            dict: The metadata of the file.
        """
        if self._metadata is None:
            self._metadata = self.get_file_metadata()
        return self._metadata

    @property
    def name(self) -> str:
        """
        The name of the file.

        Returns:
            str: The name of the file.
        """
        return self.path.name

    @property
    def extension(self) -> str:
        """
        The extension of the file.

        Returns:
            str: The extension of the file.
        """
        return self.path.suffix

    @property
    def size(self) -> int:
        """
        The size of the file in bytes.

        Returns:
            int: The size of the file in bytes.
        """
        return self.path.stat().st_size

    @property
    def created(self) -> datetime.datetime:
        """
        The date and time the file was created.

        Returns:
            datetime.datetime: The date and time the file was created.
        """
        return datetime.datetime.fromtimestamp(self.path.stat().st_ctime)

    @property
    def modified(self) -> datetime.datetime:
        """
        The date and time the file was last modified.

        Returns:
            datetime.datetime: The date and time the file was last modified.
        """
        return datetime.datetime.fromtimestamp(self.path.stat().st_mtime)

    @property
    def is_redacted(self) -> bool:
        """
        Whether the file is redacted.

        Returns:
            bool: Whether the file is redacted.
        """
        return self.parent.dataset_action["action"] == "redact" and re.search(
            self.parent.dataset_action["pattern"], self.name
        )

    @property
    def action(self) -> str:
        """
        The action to take on the file.

        Returns:
            str: The action to take on the file.
        """
        if self.is_excluded:
            return "exclude"
        if self.is_redacted:
            return "redact"
        return "include"

    @property
    def is_excluded(self) -> bool:
        """
        Whether the file is excluded.

        Returns:
            bool: Whether the file is excluded.
        """

        return bool(
            self.parent.dataset_action["action"] == "exclude"
            and re.search(self.parent.dataset_action["pattern"], self.name)
        )

    def get_file_metadata(self) -> dict:
        """
        Returns the metadata of the file.

        Returns:
            dict: The metadata of the file.
        """
        file_name = (
            f"XXXXXXXX{Path(self.name).suffix}" if self.is_redacted else self.name
        )

        return {
            "File Name": file_name,
            "Parent Directory": self.parent.name,
            "Extension": self.extension,
            "Action": self.action,
            "File Size (bytes)": self.size,
            "Date Created": self.created.strftime(self.DATE_FORMAT),
            "Last Modified": self.modified.strftime(self.DATE_FORMAT),
        }

    def as_record(self) -> dict:
        """
        Returns the file metadata as a dictionary record intended for creating a pandas
        DataFrame record. The record contains the metadata of the file and the metadata of its
        parent directory, flattened into a single dictionary.

        Returns:
            dict: The file metadata as a record.
        """
        logger.debug("Creating record for file '%s", self.path)
        return {**self.parent.metadata.flat, **self.file_metadata}



"""
This file contains the Person class, which is used to represent a person from the metadata.
"""

from dataclasses import dataclass, field
import logging


logger = logging.getLogger(__name__)

PERSON_KEYS = ["member", "maintainer", "creator", "contributor", "editor", "reviewer"]


def find_people_in_dict(d: dict) -> list["Person"]:
    """
    Given a dictionary, find all people in the dictionary and return them as a list of Person
    objects.

    Args:
        d (dict): The dictionary to search for people in.

    Returns:
        list[Person]: A list of Person objects found in the dictionary.
    """

    def _create_person(person_dict: dict) -> "Person":
        # remove the @type key if it exists
        person_dict.pop("@type", None)

        try:
            return Person(**person_dict)
        except TypeError as e:
            logger.debug("Could not create person from dictionary: %s", e)
            return None

    people = []
    for key, value in d.items():
        if isinstance(key, str) and key.lower() in PERSON_KEYS:
            if isinstance(value, list):
                for person in value:
                    if isinstance(person, dict):
                        people.append(_create_person(person))
            elif isinstance(value, dict):
                if isinstance(person, dict):
                    people.append(_create_person(person))

        if isinstance(value, dict):
            people.extend(find_people_in_dict(value))

        if isinstance(value, list):
            for item in value:
                people.extend(find_people_in_dict(item))

    return [person for person in people if person is not None]


@dataclass
class Person:
    """
    A Dataclass for representing a person from the metadata.
    """

    # We are ignoring the C103 rule for the following fields because these names are taken
    # from the schema.org Person class, and we want to keep the same names for consistency.

    name: str = field(default=None)
    givenName: str = field(default=None)  # pylint: disable=invalid-name
    familyName: str = field(default=None)  # pylint: disable=invalid-name
    email: str = field(default=None)
    jobTitle: str = field(default=None)  # pylint: disable=invalid-name
    description: str = field(default=None, repr=False)
    memberOf: str = field(default=None, repr=False)  # pylint: disable=invalid-name

    @property
    def full_name(self) -> str:
        """
        Return the full name of the person.

        If the name is set, it will be returned. If not, the givenName and familyName will be
        combined to form the full name. If neither the name nor the givenName and familyName are
        set, None will be returned.
        """
        if self.name:
            return self.name

        if self.givenName or self.familyName:
            return f"{self.givenName} {self.familyName}"

        return None

    def as_dict(self) -> dict:
        """
        Return a dictionary representation of the person.
        """
        return_dict = {"Name": self.full_name}
        if self.email:
            return_dict["Email"] = self.email
        if self.jobTitle:
            return_dict["Job Title"] = self.jobTitle
        if self.description:
            return_dict["Description"] = self.description
        if self.memberOf:
            return_dict["Member Of"] = self.memberOf
        return return_dict



"""
This file contains the Thing class, which is used to represent a thing with type and content.

A thing is anything that is defined in the schema.org vocabulary. It is represented as a
dictionary, with keys corresponding to the names of the attributes of the thing. Accessing an
attribute of the thing returns the value of the corresponding key in the content dictionary.
"""

from dataclasses import dataclass, field
import logging

logger = logging.getLogger(__name__)


@dataclass
class Thing(dict):
    """
    Represents a thing with type and content.

    Content is stored as a dictionary, with keys corresponding to the names of the attributes of
    the thing. Accessing an attribute of the thing returns the value of the corresponding key in
    the content dictionary.

    Attributes:
        type (str): The type of the thing.
        content (dict): The content of the thing.

    Methods:
        children: Returns a list of child things.
        supplement: Supplements the content of the thing with the content of another thing.
    """

    type: str = field(default="Thing")
    content: dict = field(default_factory=dict)

    def __post_init__(self):
        for key, value in self.content.items():
            setattr(self, key.replace(" ", "_"), value)

    def __getitem__(self, __key):
        if __key == "type":
            return self.type
        return self.content[__key]

    def children(self, recursive=False) -> list["Thing"]:
        """
        Returns a list of child things. If recursive is True, also includes nested children.

        If the content of the thing contains a key with @type, the value of that key is used as the
        type of the child thing. Otherwise, the type of the child thing is set to "Thing".

        Args:
            recursive (bool): Whether to include nested children.

        Returns:
            list[Thing]: A list of child things.
        """
        children = [
            Thing(type=item.get("@type", "Thing"), content=item)
            for value in self.content.values()
            if isinstance(value, list)
            for item in value
        ]

        if recursive:
            for child in children.copy():
                children.extend(child.children(recursive=True))

        return children

    def supplement(self, other: "Thing") -> None:
        """
        Supplements the content of the thing with the content of another thing.

        Will not overwrite existing keys, but will add new keys that do not exist in this thing.
        Will also check all nested children and supplement them recursively.

        Args:
            other (Thing): The other thing.
        """
        logger.debug("Supplementing with other thing: %s", other)

        for key, value in other.content.items():
            if key not in self.content:
                logger.debug("Adding key: %s, value: %s", key, value)
                self.content[key] = value
            elif isinstance(value, list):
                logger.debug("Extending self.content[%s] with %s", key, value)
                self.content[key].extend(value)
            elif isinstance(value, dict):
                for child in self.content[key]:
                    child.supplement(Thing(content=value))
        logger.info("Finished supplementing, self.content is now: %s", self.content)



"""
This Script is used to generate a demo project directory structure for testing the Self Documenting
Data Management Plan (SDDMP) tool. The script takes a directory structure as input and creates
the directories and subdirectories in the output directory. The input can be provided as a text file
or as a string pasted into the terminal. The output directory is created if it does not exist.

The script can be run from the command line using the following command:
```
python -m sddmp.demo -o example_project -i example_structure.txt
```
"""

import argparse
import logging
from pathlib import Path
import shutil
import sys

logger = logging.getLogger(__name__)


def _parse_cli():
    parser = argparse.ArgumentParser(
        description="Create a demonstration project directory structure for testing."
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="example_project",
        help="The output directory to create and write the demonstration project to.",
    )
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        help=(
            "A Text file containing a directory structure to build the demonstration project "
            "from."
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Print verbose output.",
    )
    parser.add_argument(
        "--dry_run",
        action="store_true",
        help="Print the output without writing to the file system.",
    )
    return parser.parse_args()


def create_directory_structure(base_path, structure_string: str, dry_run=False):
    """
    Create a directory structure from a string representation.

    Args:
        base_path (str): The base path to create the directory structure in.
        structure_string (str): A string representation of the directory structure to create.
        dry_run (bool): If True, the directory structure will not be created.

    Returns:
        None
    """
    lines = structure_string.split("\n")
    path_stack = []
    for line in lines:
        if line.strip() == "":
            continue
        depth = line.index(line.lstrip()) // 3
        name = line.lstrip()
        while len(path_stack) > depth:
            path_stack.pop()
        path_stack.append(name)
        path = Path(base_path, *path_stack)
        log_func = logger.info if dry_run else logger.debug
        log_func("Creating %s", path)
        if not dry_run:
            path.mkdir(parents=True, exist_ok=True)


if __name__ == "__main__":
    args = _parse_cli()

    # Set up logging.
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    print(f"Creating example project at {args.output}")

    structure = ""
    if args.input:
        print(f"Using input file {args.input}")
        with open(args.input, encoding="utf-8") as f:
            structure = f.read()
    else:
        print(
            "Please paste a directory structure into the terminal here (Press Enter to finish):"
        )
        while True:
            text = input()
            if text == "":
                break
            structure += text + "\n"

    # Let the user see the structure that will be created.
    print("The following directory structure will be created:")
    print(structure)
    if input("Is this correct? (y/n): ").lower() != "y":
        print("Exiting.")
        sys.exit()

    create_directory_structure(args.output, structure, args.dry_run)

    if not args.dry_run:
        # Drop a copy of the example README in the output directory.
        readme = Path(__file__).parent / "resources/README_example.yaml"
        readme_path = Path(args.output, "README.yaml")
        shutil.copy(readme, readme_path)

    print("Done!")



"""
This module contains the Directory class, which is used to represent a directory in the file
system. The Directory class is used to create a tree structure of directories and files, and to
create a dataframe of all files in the directory tree.
"""

from dataclasses import dataclass, field
import logging
from pathlib import Path

import pandas as pd
import yaml

from .file import File
from .filetree import FileTree
from ..metadata import Metadata

logger = logging.getLogger(__name__)


@dataclass
class Directory:
    """
    Represents a directory in the file system.

    Attributes:
        path (Path): The path to the directory.
        parent (Directory): The parent directory of this directory.
        children (list[Directory]): The child directories of this directory.
        files (list[File]): The files in this directory.
        metadata_filename (str): The name of the metadata file in this directory.

    Properties:
        name (str): The name of the directory.
        metadata (Metadata): The metadata of the directory.
        all_descendants (list[Directory]): All descendant directories of this directory.

    Methods:
        get_metadata: Load the metadata file in this directory, or create a new metadata object.
        as_dataframe: Create a dataframe of all files in the directory tree.
    """

    path: Path
    parent: "Directory" = None
    children: list["Directory"] = field(default_factory=list, repr=False)
    files: list[File] = field(default_factory=list, repr=False)
    metadata_filename = "README.yaml"

    _filetree: FileTree = field(default=None, repr=False)

    @property
    def name(self):
        """
        The name of the directory.
        """
        return self.path.name

    @property
    def metadata(self):
        """
        The metadata of the directory.
        """
        if self._metadata is None:
            self._metadata = self.get_metadata()
        return self._metadata

    @property
    def all_descendants(self):
        """
        All descendant directories of this directory.
        """
        descendants = [self]
        for child in self.children:
            descendants.append(child)
            descendants.extend(child.all_descendants)
        return descendants

    @property
    def filetree(self) -> str:
        """
        Get the filetree of the directory as a string.
        """
        return self._filetree.as_string()

    @property
    def dataset_action(self) -> dict[str, str]:
        """
        There is a very specific metadata attribute we can look for that the user can update to
        specify what action should be taken on the data in this directory.

        Will be one of:
        - "include"
        - "exclude"
        - "redact"

        The reuturned dictionary will be in the form:
        {
            "action": "include" | "exclude" | "redact",
            "pattern": "regex pattern to match files to apply action to" | None
        }

        Returns:
            dict[str, str]: The action to take on the dataset in the directory.
        """
        if (
            "Dataset" not in self.metadata
            or "potentialAction" not in self.metadata["Dataset"]
        ):
            return {
                "action": "include",
                "pattern": ".*",
            }

        action = self.metadata["Dataset"].get("potentialAction", [{"name": "include"}])
        return {
            "action": action[0]["name"],
            "pattern": action[0].get("pattern", ".*"),
        }

    def __post_init__(self):
        # If the directory has no parent, it is the root directory.
        if self.parent is None:
            self.parent = self

        # If the directory has no children, it is a leaf directory.
        if self.children is None:
            self.children = []

        # If the directory has no files, it is an empty directory.
        if self.files is None:
            self.files = []

        self._metadata = None

        self._filetree = FileTree(self.path)

        logger.debug("Created directory object %s", self.path)

    @property
    def file_dataframe(self) -> pd.DataFrame:
        df = pd.DataFrame([file.get_file_metadata() for file in self.files])
        full_df = pd.concat([df] + [child.file_dataframe for child in self.children])
        return full_df

    @property
    def num_excluded_files(self) -> int:
        if "Action" not in self.file_dataframe:
            return 0

        return sum(self.file_dataframe["Action"] == "exclude")

    def get_metadata(self) -> Metadata:
        """
        Get the metadata of the directory.

        If the metadata file exists, load it. Otherwise, create a new metadata object. If this
        directory is not the root directory, supplement the metadata with the metadata of the
        parent directory.

        Returns:
            Metadata: The metadata of the directory.
        """
        # Create a path to the metadata file in this directory.
        metadata_path = self.path / self.metadata_filename

        # Load the metadata file if it exists, otherwise create a new metadata object.
        my_metadata = (
            Metadata.load(metadata_path) if metadata_path.exists() else Metadata()
        )

        # If this directory is not the root directory, supplement the metadata with the
        if self.parent != self:
            my_metadata.supplement(self.parent.metadata)

        return my_metadata

    def as_dataframe(self) -> pd.DataFrame:
        """
        Create a dataframe of all files in the directory tree.

        Returns:
            pd.DataFrame: A dataframe of all files in the directory tree.
        """
        logger.debug("Creating dataframe for directory %s", self.path)
        # Create a dataframe by collecting all files in this directory as records.
        df = pd.DataFrame([file.as_record() for file in self.files])

        # Extend the dataframe by collecting all files in the child directories as records.
        df = pd.concat([df] + [child.as_dataframe() for child in self.children])

        return df

    def metadata_as_plaintext(self) -> dict[str, str]:
        """
        Create a dictionary of the directory tree in plaintext.

        Keys are the root items in the metadata

        Returns:
            dict[str, str]: A dictionary of the directory tree in plaintext.
        """
        logger.debug("Creating plaintext for directory %s", self.path)

        return_dict = {}
        for key, value in self.get_metadata().items():
            return_dict[key] = yaml.dump(value, default_flow_style=False)

        return return_dict

    def file_records_dataframe(self) -> pd.DataFrame:
        """
        Create a dataframe of all files in the directory tree.

        Returns:
            pd.DataFrame: A dataframe of all files in the directory tree.
        """
        full_df = self.file_dataframe.copy()

        # Strip out any records where the Action is "exclude"
        if "Action" in full_df.columns:
            full_df = full_df[full_df["Action"] != "exclude"]

        return full_df



"""
This file contains the SchemaOrgValidator class, which is used to validate terms against the
schema.org vocabulary.
"""

import tempfile
import logging
from pathlib import Path
import time

import rdflib
import requests

logger = logging.getLogger(__name__)


class Singleton(type):
    """
    This class is a metaclass for creating singleton classes. It ensures that only one instance of
    the class is created.

    Attributes:
        _instances (dict): A dictionary of instances of the class.

    Methods:
        __call__: Returns the instance of the class if it already exists, otherwise creates a new
            instance and returns it.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            logger.debug("Initializing SchemaOrgValidator object")
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class SchemaOrgValidator(metaclass=Singleton):
    """
    This class is used to validate terms against the schema.org vocabulary.

    Attributes:
        GITHUB_URL (str): The URL of the schema.org file on GitHub.
        SUBDIRECTORY (str): The subdirectory in the cache directory where the schema.org file is
            stored.
        FILE_NAME (str): The name of the schema.org file.

    Properties:
        cache_file_path (Path): The file path of the cache file.
        all_terms (set): A set of all the terms in the schema.org file.

    Methods:
        __init__: Initializes the SchemaOrgValidator object.
        _check_and_download: Checks if the schema.org file exists in the cache directory and
            downloads it if not.
        get_term_details: Returns the details of a term in the schema.org file.
        term_exists: Checks if a given term exists in the graph.
    """

    _instance = None

    GITHUB_URL = (
        "https://github.com/schemaorg/schemaorg/raw/main/data/releases/23.0/"
        "schemaorg-all-http.rdf"
    )
    SUBDIRECTORY = "sddmp"
    FILE_NAME = "schemaorg-all-http.rdf"

    @property
    def cache_file_path(self):
        """
        Returns the file path of the cache file.

        Returns:
            pathlib.Path: The file path of the cache file.
        """
        return self.temp_dir.joinpath(self.SUBDIRECTORY, self.FILE_NAME)

    @property
    def all_terms(self) -> set:
        """
        Gets a set of all the terms in the schema.org file.

        Returns:
            set: A set of all the terms in the schema.org file.
        """
        if self._all_terms is None:
            self._all_terms = set()
            for _, _, object_ in self.g:
                # Only add terms that are not single words and are not part of a URL
                clean_object = str(object_).replace(r"http://schema.org/", "")

                if len(str(clean_object).split(" ")) == 1:
                    self._all_terms.add(str(clean_object))
        return self._all_terms

    def __init__(self, temp_dir: Path = None):
        logger.debug("Initializing SchemaOrgValidator object")
        self.temp_dir = (
            Path(tempfile.gettempdir()) if temp_dir is None else Path(temp_dir)
        )
        self.cache_file_path.parent.mkdir(exist_ok=True)
        self._check_and_download()
        self.g = rdflib.Graph()
        self.g.parse(self.cache_file_path)

        self._all_terms = None
        logger.debug("Initialized SchemaOrgValidator object")
        self._initialized = True

    def _check_and_download(self):
        """
        Checks if the schema.org file exists in the cache directory and downloads it if not.

        Raises:
            requests.exceptions.RequestException: If there is an error while downloading the
            schema.org file.
        """
        if not self.cache_file_path.exists():
            start = time.time()
            logger.info("Downloading schema.org file...")
            try:
                response = requests.get(self.GITHUB_URL, timeout=20)
                response.raise_for_status()
                with open(self.cache_file_path, "wb") as f:
                    f.write(response.content)
                logger.info(
                    "Downloaded schema.org file in %.2f seconds", time.time() - start
                )
            except requests.exceptions.RequestException as err:
                logger.error("Failed to download schema.org file: %s", err)
                raise err

    def get_term_details(self, term: str) -> dict:
        """
        Get the details of a term in the schema.org file.

        Args:
            term (str): The term to get the details of.

        Returns:
            dict: The details of the term in the schema.org file.
        """
        logger.debug("Getting details for term '%s'", term)
        for subject, predicate, object_ in self.g:
            if term in [str(subject), str(predicate), str(object_)]:
                logger.debug("Found details for term '%s'", term)
                return {
                    "subject": subject,
                    "predicate": predicate,
                    "object": object_,
                }
        logger.debug("Did not find details for term '%s'", term)
        return {}

    def term_exists(self, term: str) -> True:
        """
        Checks if a given term exists in the graph.

        Args:
            term (str): The term to check.

        Returns:
            bool: True if the term exists in the graph, False otherwise.
        """
        return term in self.all_terms



# pylint: disable=too-many-arguments

"""
This file contains the PiePlot class, which is used to create a pie plot using the Plotly library.
"""

import plotly.graph_objects as go

from .base_plot import BasePlot


class PiePlot(BasePlot):
    """
    Represents a pie plot using the Plotly library.

    Attributes:
        title (str): The title of the plot.
        labels (list[str]): The labels of the pie plot.
        values (list[float]): The values of the pie plot.

    Methods:
        get: Get the pie plot as a Plotly graph object figure.
    """

    def __init__(
        self,
        title,
        labels,
        values,
        text=None,
        hovertemplate="%{label}: %{value} (%{percent})<extra></extra>",
    ):
        super().__init__(title=title)
        self.labels = labels
        self.values = values
        self.text = text
        self.hovertemplate = (
            hovertemplate if text is None else hovertemplate + "<br>%{text}"
        )

    def get(self) -> go.Figure:
        fig = (
            super()
            .get()
            .add_trace(
                go.Pie(
                    labels=self.labels,
                    values=self.values,
                    text=self.text,
                    hovertemplate=self.hovertemplate,
                )
            )
        )
        fig.update_traces(textposition="inside")
        fig.update_layout(uniformtext_minsize=12, uniformtext_mode="hide")
        return fig



"""
This file represents the file system. It contains the FileSystem class, which is used to read a
directory and create a tree of Directory and File objects.
"""

from pathlib import Path
from .directory import Directory
from .file import File


class FileSystem:
    """
    A class for representing the file system.

    Methods:
        read_directory: Reads a directory and creates a tree of Directory and File objects.
    """

    def read_directory(self, path: Path, parent: Directory = None) -> Directory:
        """
        Reads a directory and creates a tree of Directory and File objects.

        Applies recursively to all subdirectories, resulting in a tree of Directory and File
        objects.

        Args:
            path (Path): The path to the directory to read.
            parent (Directory): The parent directory of the directory to read.

        Returns:
            Directory: The directory object representing the directory and its contents.
        """
        # In case the path is provided as a string, convert it to a Path object.
        if isinstance(path, str):
            path = Path(path)

        directory = Directory(path=path, parent=parent)
        for child in path.iterdir():
            if child.is_file():
                directory.files.append(File(path=child, parent=directory))
            elif child.is_dir():
                directory.children.append(self.read_directory(child, directory))
        return directory

    def get_directory(self, path: Path) -> Directory:
        """
        Returns a Directory object representing the directory and its contents.

        Args:
            path (Path): The path to the directory.

        Returns:
            Directory: The directory object representing the directory and its contents.
        """
        return self.read_directory(path)

    def get_file(self, path: Path) -> File:
        """
        Returns a File object representing the file.

        Args:
            path (Path): The path to the file.

        Returns:
            File: The file object representing the file.
        """
        return File(path=path, parent=Directory(path.parent))

    def get_directory_structure(self, directory: Directory) -> dict:
        """
        Get a nested list of all directories in the file system.

        Args:
            directory (Directory): The root directory to start from.

        Returns:
            dict: A dictionary representing the directory structure.
        """
        sorted_children = sorted(directory.children, key=lambda child: child.path.name)

        structure = {
            "name": directory.path.name,
            "children": [
                self.get_directory_structure(child) for child in sorted_children
            ],
        }
        return structure



"""
This file contains the PlotGenerator class, which is used to generate plots from a dataframe for
use in the HTML page.
"""

import logging

from .pie_plot import PiePlot

logger = logging.getLogger(__name__)


class PlotGenerator:
    """
    Class for generating plots from a dataframe for use in the HTML page.
    """

    def __init__(self, df):
        logger.debug("Creating a PlotGenerator object.")
        logger.debug("Dataframe columns: %s", df.columns)
        self.df = df

    def generate_pie_plot(self, column: str) -> PiePlot:
        """
        Generates a pie plot for a column in the dataframe.

        Args:
            df (pd.DataFrame): The dataframe.
            column (str): The name of the column in the dataframe.

        Returns:
            PiePlot: The pie plot.
        """
        # Make a copy of the dataframe
        df = self.df.copy()

        # Create a dataframe that counts the number of files with each extension.
        df = df[column].value_counts().reset_index()

        # Rename the columns.
        df.columns = [column, "count"]

        # Create a pie plot.
        plot = PiePlot(
            title=f"Number of files by {column}".replace("-", " ").title(),
            labels=df[column],
            values=df["count"],
        )

        return plot

    def all_plots(self) -> dict:
        """
        Generates all of the pie plots used in the HTML page.

        returns:
            dict: A dictionary of the pie plots.
        """
        # If we have a 0 length dataframe, return an empty dictionary.
        if len(self.df) == 0:
            return {}

        plots = {}

        # Generate pie plots for the file extensions
        file_extension_counts = self.df["Extension"].value_counts()
        plots["file_extension"] = PiePlot(
            "File Extensions",
            file_extension_counts.index,
            file_extension_counts.values,
            text=[
                f"{count} {extension_name} files"
                for extension_name, count in file_extension_counts.items()
            ],
        )

        size_by_extension = self.df.groupby("Extension")["File Size (bytes)"].sum()
        if len(size_by_extension) > 0:
            text = [
                f"{extension_name}: {size / 1024**2:.2f} MB"
                for extension_name, size in size_by_extension.items()
            ]

            plots["file_size_by_extension"] = PiePlot(
                "File Size by Extension",
                size_by_extension.index,
                size_by_extension.values,
                text=text,
                hovertemplate="%{value} Bytes (%{percent})<extra></extra>",
            )

        size_by_directory = self.df.groupby("Parent Directory")[
            "File Size (bytes)"
        ].sum()
        if len(size_by_directory) > 0:
            text = [
                f"{size / 1024**2:.2f} MB"
                for extension_name, size in size_by_directory.items()
            ]

            plots["file_size_by_directory"] = PiePlot(
                "File Size by Directory",
                size_by_directory.index,
                size_by_directory.values,
                text=text,
            )

        return plots



# pylint: disable=too-few-public-methods

"""
This file contains the ReferencePage class, which is used to generate an HTML page for the
glossary of terms and special keys used in the project.
"""

from pathlib import Path

from .html_page import HTMLPage


class ReferencePage(HTMLPage):
    """
    This is a class for generating an HTML page for the glossary of terms and special keys used in
    the project. Presents similar information to the project README file, but will be included in
    the output files.
    """

    def __init__(self, input_directory: Path, output_directory: Path):
        super().__init__("reference.html.j2", output_directory)
        self.input_directory = input_directory

    def get_relative_path_to_root(self) -> str:
        return ".."

    def get_path(self) -> Path:
        return (
            Path(self.output_directory) / Path(self.input_directory) / "reference.html"
        )

    def get_depth(self) -> int:
        return 0



"""
This file contains the DirectoryPage class, which is used to generate an HTML page for a directory
in the file system.
"""

from pathlib import Path

from ..filesystem import Directory
from .html_page import HTMLPage
from .plots.plot_generator import PlotGenerator


class DirectoryPage(HTMLPage):
    """
    Represents an HTML page for a directory in the file system.

    Attributes:
        directory (Directory): The directory for which the HTML page is to be generated.
        output_directory (Path): The directory in which the HTML page is to be saved.

    Methods:
        generate: Generates the HTML page for the directory.
        generate_pie_plot: Generates a pie plot for a column in the dataframe.
    """

    def __init__(self, directory: Directory, output_directory: Path):
        super().__init__("directory.html.j2", output_directory)
        self.directory = directory

    def get_relative_path_to_root(self) -> str:
        """
        Returns the relative path from the output directory to the root directory.
        """
        num_path_parts = len(self.directory.path.parts)
        if num_path_parts == 1:
            return ".."

        return "../" * (num_path_parts)

    def get_path(self) -> Path:
        """
        Returns the path to the HTML file.
        """
        return self.output_directory / self.directory.path / "index.html"

    def get_depth(self) -> int:
        """
        Returns the depth of the page in the directory structure.
        """
        return len(self.directory.path.parts)

    def generate(self) -> None:
        """
        Generates the HTML page for the directory and saves it in the output directory.
        """
        file_records_df = self.directory.file_records_dataframe()

        plot_generator = PlotGenerator(file_records_df)

        # Render the template with the dataframe.
        super().generate(
            project_title=self.directory.metadata["ResearchProject"]["name"],
            my_path=self.directory.path.as_posix(),
            num_files=len(file_records_df),
            num_excluded_files=self.directory.num_excluded_files,
            num_directories=len(self.directory.all_descendants) + 1,
            people=self.directory.metadata.get_people(),
            filetree=self.directory.filetree,
            plots={
                name: plot.as_json()
                for name, plot in plot_generator.all_plots().items()
            },
            metadata=self.directory.metadata_as_plaintext(),
            file_records_df=file_records_df,
        )



"""
This file contains the BasePlot class, which is the base class for all the plots in the package.
"""

import json

import plotly

from .palette import Palette


class BasePlot:
    """
    The BasePlot class is the base class for all the plots in the package. It contains the basic
    attributes and methods that are common to all the plots.

    Attributes:
        title (str): The title of the plot.
        title_font (dict): The font of the title of the plot.
            (default: {"family": "Arial", "size": 24})
        xaxis_title (str): The title of the x-axis of the plot. (default: None)
        yaxis_title (str): The title of the y-axis of the plot. (default: None)

    Properties:
        plot_bgcolor (str): The background color of the plot.
            (default: Palette.COSCINE_LIGHT_BLUE2)
        paper_bgcolor (str): The background color of the paper.
            (default: Palette.COSCINE_WHITE)

    Methods:
        get: Returns the plot as a plotly.graph_objects.Figure object.
        as_json: Returns the plot as a dictionary that can be written to a JSON file.
    """

    def __init__(
        self,
        title: str,
        title_font: dict = None,
        xaxis_title: str = None,
        yaxis_title: str = None,
    ):
        self._figure = plotly.graph_objects.Figure()
        self.title = title

        self.title_font = title_font
        if self.title_font is None:
            self.title_font = {"family": "Arial", "size": 24}

        self.xaxis_title = xaxis_title
        self.yaxis_title = yaxis_title
        self.plot_bgcolor = str(Palette.COSCINE_LIGHT_BLUE2)
        self.paper_bgcolor = str(Palette.COSCINE_WHITE)

    def get(self) -> plotly.graph_objects.Figure:
        """
        Get the plot as a plotly.graph_objects.Figure object.

        Returns:
            plotly.graph_objects.Figure: The plot as a plotly.graph_objects.Figure object.
        """
        return self._figure.update_layout(
            title=self.title,
            title_font=self.title_font,
            xaxis_title=self.xaxis_title,
            yaxis_title=self.yaxis_title,
            plot_bgcolor=self.plot_bgcolor,
            paper_bgcolor=self.paper_bgcolor,
        )

    def as_json(self) -> dict:
        """
        Get the plot as a dictionary that can be written to a JSON file.

        Returns:
            dict: The plot as a dictionary that can be written to a JSON file.
        """
        return json.dumps(self.get(), cls=plotly.utils.PlotlyJSONEncoder)



"""
This file contains the HTMLPage class, which is used as a base class for generating HTML pages.

This class is not intended to be used directly, but rather to be subclassed by classes that
generate specific types of HTML pages.
"""

from abc import ABC, abstractmethod
from pathlib import Path
import pkg_resources

from jinja2 import Environment, FileSystemLoader


class HTMLPage(ABC):
    """
    This is an abstract base class containing common methods and properties for generating HTML
    pages.
    """

    # Class Variable for storing the directory structure
    directory_structure = None

    def __init__(self, template_name: str, output_directory: Path):
        self.output_directory = output_directory
        template_dir = pkg_resources.resource_filename(__name__, "/templates")
        self.jinja_env = Environment(
            loader=FileSystemLoader(template_dir), autoescape=True
        )
        self.template = self.jinja_env.get_template(template_name)

    @abstractmethod
    def get_relative_path_to_root(self) -> str:
        """
        Returns the relative path from the this page's path to the root directory.

        Used to locate static files in the output directory.

        Returns:
            str: The relative path from this page's path to the root directory.
        """

    @abstractmethod
    def get_path(self) -> Path:
        """
        Returns the path to the HTML file.

        Returns:
            str: The path to the HTML file.
        """

    @abstractmethod
    def get_depth(self) -> int:
        """
        Returns the depth of the page in the directory structure.

        TODO: I honestly can't remember why we have both this and relative path to root
        """

    def generate(self, **kwargs) -> None:
        """
        Generate an HTML page.
        """
        html = self.template.render(
            directory_structure=self.directory_structure,
            relative_path_to_root=self.get_relative_path_to_root(),
            depth=self.get_depth(),
            **kwargs
        )

        # Create the directory for the html file if it does not exist.
        Path(self.get_path()).parent.mkdir(parents=True, exist_ok=True)

        # Write the html file.
        with open(self.get_path(), "w", encoding="utf-8") as f:
            f.write(html)



"""
This file contains the Palette class, which is used to define the color palette used in the plots.
"""

from enum import Enum


class Palette(Enum):
    """
    Represents the color palette used in the plots.
    """

    COSCINE_BLACK = "#212529"
    COSCINE_GREY = "#646567"
    COSCINE_WHITE = "#ffffff"
    COSCINE_DARK_BLUE = "#00549F"
    COSCINE_LIGHT_BLUE = "#8EBAE6"
    COSCINE_LIGHT_BLUE2 = "#e8f1fa"

    def __str__(self):
        return self.value



# pylint: disable=too-few-public-methods

"""
This module contains the FileTree class, which is used to represent a directory tree in the file
system. The FileTree class is used to create a tree structure of directories as a string, similar
to the output of the `tree` command in Unix-like operating systems.
"""

import logging
from pathlib import Path

logger = logging.getLogger(__name__)


class FileTree:
    """
    This class represents a directory in the file system, and is used to create a tree structure of
    directories as a string, similar to the output of the `tree` command in Unix-like operating
    systems.
    """

    def __init__(self, directory: Path):
        self.directory = directory

    def as_string(self, level: int = 0) -> str:
        """
        Return the directory tree as a string.

        Args:
            level (int): The level of the directory in the tree.

        Returns:
            str: The directory tree as a string.
        """
        indent = "│   " * (level - 1) + "├── " if level > 0 else ""
        tree = f"{indent}{self.directory.name}\n"
        tree = ""
        children = list(self.directory.iterdir())
        for i, child in enumerate(children):
            if i == len(children) - 1:
                indent = "│   " * level + "└── "
            else:
                indent = "│   " * level + "├── "
            tree += f"{indent}{child.name}\n"
            if child.is_dir():
                child_tree = FileTree(child)
                tree += child_tree.as_string(level + 1)

        if level == 0:
            return self.directory.name + "\n" + tree

        return tree



"""
This file is used to create static html files for displaying in GitHub Pages.

Usage:
    python create_html_files.py -i data -o docs
"""

import shutil

import argparse
import logging
from pathlib import Path
import pkg_resources

from sddmp.filesystem import FileSystem
from sddmp.outputs import DirectoryPage, ReferencePage

logger = logging.getLogger(__name__)


def _parse_cli():
    parser = argparse.ArgumentParser(
        description="Create static html files for displaying in GitHub Pages."
    )
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default="data",
        help="The input directory to create html files for.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="docs",
        help="The output directory for the html files.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Print verbose output.",
    )
    parser.add_argument(
        "--source_prefix",
        default="",
        help="The prefix to add to the source path when loading the static files",
    )
    return parser.parse_args()


def create_index(root_directory, output_directory):
    """
    Create a simple index page for navigating around the html files.

    THIS IS NOT A FINAL IMPLEMENTATION. This is just a quick and dirty way to create an index page.
    """
    # Create a simple html file that contains links to all of the directories.
    html = "<ul>"
    for my_directory in root_directory.all_descendants:
        html += (
            f'<li><a href="{my_directory.path}/index.html">{my_directory.path}</a></li>'
        )
    html += "</ul>"

    # Create the path to the html file.
    my_path = Path(output_directory) / "index.html"

    # Write the html file.
    with open(my_path, "w", encoding="utf-8") as f:
        f.write(html)

    print(f"Created index file at {my_path}")


if __name__ == "__main__":
    args = _parse_cli()

    # Set up logging.
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    fs = FileSystem()
    root = fs.read_directory(args.input)

    DirectoryPage.directory_structure = fs.get_directory_structure(root)["children"]
    ReferencePage.directory_structure = fs.get_directory_structure(root)["children"]

    # Delete the output directory if it exists.
    if Path(args.output).exists():
        logger.info("Deleting existing output directory at %s", args.output)
        for path in Path(args.output).glob("*"):
            if path.is_file():
                path.unlink()
            else:
                shutil.rmtree(path)
        Path(args.output).mkdir(exist_ok=True, parents=True)

    # Move the static directory to the output directory.
    source_dir = pkg_resources.resource_filename(__name__, "/outputs/static/")
    logger.info("Copying static files from %s to %s/static", source_dir, args.output)
    shutil.copytree(source_dir, Path(args.output) / "static")

    # Create a reference page
    reference = ReferencePage(args.input, args.output)
    reference.generate()

    # Start with the root and iterate recursively through all directories.
    # For each directory, create an html file.
    for directory in root.all_descendants:
        page = DirectoryPage(directory, args.output)
        page.generate()

    # Create an index to help navigate around
    create_index(root, args.output)



"""
This file contains the Metadata class, which is used to represent metadata for a file or directory.
"""

import yaml

from .thing import Thing
from .person import Person, find_people_in_dict

# from .schema_org_validator import SchemaOrgValidator


def all_keys(d) -> list[str]:
    """
    Iterates over all keys in the provided dictionary, including nested keys.

    Args:
        d (dict): The dictionary to iterate over.

    Returns:
        list[str]: A list of all keys in the dictionary.
    """
    keys = []
    for key, value in d.items():
        keys.append(key)
        if isinstance(value, dict):
            keys.extend(all_keys(value))
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    keys.extend(all_keys(item))
    return list(set(keys))


def dict_supplement(dict1, dict2) -> dict:
    """
    Supplements the content of dict1 with the content of dict2.

    Includes nested dictionaries, lists and values recursively.

    Args:
        dict1 (dict): The dictionary to supplement.
        dict2 (dict): The dictionary to supplement with.

    Returns:
        dict: The supplemented dictionary.
    """
    for key, value in dict2.items():
        if key not in dict1:
            dict1[key] = value
        elif isinstance(dict1[key], dict) and isinstance(value, dict):
            dict1[key] = dict_supplement(dict1[key], value)
    return dict1


class Metadata(dict):
    """
    A class for representing metadata for a file or directory.

    Class Methods:
        load: Loads metadata from a file.

    Properties:
        flat: A flattened version of the metadata.

    Methods:
        supplement: Supplements the content of the metadata with the content of another metadata
            object.
        get_base_objects: Returns a list of Thing objects representing the base objects in the
            metadata.
        all_keys: Iterates over all keys in the metadata object, including nested keys.
        validate_keys: Validates the keys in the metadata object against the schema.org vocabulary.
        flatten: Flattens the metadata object into a dictionary.
    """

    @property
    def flat(self) -> dict:
        """
        Get a flattened version of the metadata.

        Returns:
            dict: The flattened metadata.
        """
        if self._flat_metadata is None:
            self._flat_metadata = self.flatten()
        return self._flat_metadata

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.schema_org_validator = None  # SchemaOrgValidator()
        self._flat_metadata = None

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            return None

    @classmethod
    def load(cls, file_path: str) -> "Metadata":
        """
        Load metadata from a file.

        Args:
            file_path (str): The file path of the metadata file.

        Returns:
            Metadata: The metadata object.
        """
        with open(file_path, "r", encoding="utf-8") as file:
            data = yaml.safe_load(file)
            return cls(data)

    @property
    def is_valid(self) -> bool:
        """
        Determines whether the metadata is valid based on the schema.org vocabulary.

        Returns:
            bool: True if the metadata is valid, otherwise False.
        """
        return True
        return all(value["valid"] for value in self.validate_keys().values())

    def supplement(self, other: "Metadata") -> None:
        """
        Supplements the content of the metadata with the content of another metadata object.

        Args:
            other (Metadata): The metadata object to supplement with.

        Returns:
            None
        """
        self.update(dict_supplement(self, other))

    def get_base_objects(self) -> list[Thing]:
        """
        Get a list of Thing objects representing the base objects in the metadata.

        Returns:
            list[Thing]: A list of Thing objects representing the base objects in the metadata.
        """

        objects = []
        for key, value in self.items():
            objects.append(Thing(type=key, content=value))
        return objects

    def get_people(self) -> list[Person]:
        """
        Get a list of Person objects found in the metadata.

        Returns:
            list[Person]: A list of Person objects found in the metadata.
        """
        return find_people_in_dict(self)

    def all_keys(self) -> list[str]:
        """
        Iterates over all keys in the metadata object, including nested keys.

        Returns:
            list[str]: A list of all keys in the metadata object.
        """
        # Initialize an empty list to store the keys
        return all_keys(self)

    def validate_keys(self) -> dict[str, bool]:
        """
        Validates the keys in the metadata object against the schema.org vocabulary.

        Returns:
            dict[str, bool]: A dictionary of keys and their validation status.
        """
        return {
            key: {
                **{"valid": self.schema_org_validator.term_exists(key)},
                **self.schema_org_validator.get_term_details(key),
            }
            for key in self.all_keys()
        }

    def flatten(self, parent_key="", sep="_") -> dict:
        """
        Flatten the metadata object into a dictionary.

        Args:
            parent_key (str): The parent key to use when flattening nested dictionaries.
            sep (str): The separator to use when concatenating keys.

        Returns:
            dict: The flattened metadata.
        """
        # Create an empty list to store the flattened key-value pairs
        items = []

        # Iterate over each key-value pair in the Metadata object
        for key, value in self.items():
            # Create a new key by concatenating the parent key and the current key, separated by
            # the provided separator
            new_key = f"{parent_key}{sep}{key}" if parent_key else key

            # If the value is a dictionary, recursively call this method on the value
            if isinstance(value, dict):
                items.extend(
                    Metadata(value).flatten(parent_key=new_key, sep=sep).items()
                )

            # If the value is a list of dictionaries, iterate over each element and recursively
            # call this method on each element
            elif isinstance(value, list) and all(isinstance(i, dict) for i in value):
                for i, elem in enumerate(value):
                    items.extend(
                        Metadata(elem)
                        .flatten(parent_key=f"{new_key}{sep}{i}", sep=sep)
                        .items()
                    )

            # If the value is neither a dictionary nor a list of dictionaries, add the key-value
            # pair to the items list
            else:
                items.append((new_key, value))

        # Convert the list of key-value pairs to a dictionary and return it
        return dict(items)
